#include "common.h"
#include <cds/misc/chase-lev-deque.h>
#include <cds_test/stress_test.h>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <memory>
#include <thread>

using namespace std;

namespace {

typedef cds_others::ChaseLevDeque Deque;
static size_t s_nDequeStealerThreadCount = 3;
static size_t s_nDequeMainPassCount = 100000000;

class ChaseLevDequeTest : public cds_test::stress_fixture {
protected:
  static Deque *deque;
  static atomic_int terminate_stealer;
  static ullong *sums;
  static ullong *succ_counts;
  static ullong push_sum;
  static ullong push_count;

  static void SetUpTestCase() {
    cds_test::config const &cfg = get_config("Misc");
    GetConfig(DequeStealerThreadCount);
    GetConfig(DequeMainPassCount);
  }

  static void StealerThread(int index) {
    while (!terminate_stealer.load(memory_order_relaxed)) {
      while (true) {
        int res = deque->steal();
        if (res != EMPTY && res != ABORT) {
          sums[index] += res;
          succ_counts[index]++;
        } else {
          break;
        }
      }
    }
  }

  static void MainThread(int index, int push_percentage) {
    for (ullong i = 0; i < s_nDequeMainPassCount; i++) {
      if ((::rand() % 100) < push_percentage) {
        int item = ::rand() % 100;
        deque->push(item);
        push_sum += item;
        push_count++;
      } else {
        int res = deque->take();
        if (res != EMPTY) {
          sums[index] += res;
          succ_counts[index]++;
        }
      }
    }
    while (true) {
      int res = deque->take();
      if (res != EMPTY) {
        sums[index] += res;
        succ_counts[index]++;
      } else {
        break;
      }
    }
  }
};

atomic_int ChaseLevDequeTest::terminate_stealer;
ullong *ChaseLevDequeTest::sums;
ullong *ChaseLevDequeTest::succ_counts;
ullong ChaseLevDequeTest::push_count;
ullong ChaseLevDequeTest::push_sum;
Deque *ChaseLevDequeTest::deque;

TEST_F(ChaseLevDequeTest, DequePushPopTake) {
  deque = new Deque();
  push_sum = 0;
  sums = (ullong *)calloc(1, sizeof(ullong) * (s_nDequeStealerThreadCount + 1));
  succ_counts =
      (ullong *)calloc(1, sizeof(ullong) * (s_nDequeStealerThreadCount + 1));
  srand(time(NULL));

  // Stealer threads
  std::unique_ptr<std::thread[]> threads(
      new std::thread[s_nDequeStealerThreadCount]);
  for (ullong i = 0; i < s_nDequeStealerThreadCount; i++) {
    threads[i] = std::thread(StealerThread, i);
  }

  for (int i = 90; i > 0; i -= 10) {
    MainThread(s_nDequeStealerThreadCount, i);
  }

  terminate_stealer.store(1, memory_order_relaxed);
  for (ullong i = 0; i < s_nDequeStealerThreadCount; i++) {
    threads[i].join();
  }

  // Result analysis
  ullong received_sum = 0;
  ullong overall_count = 0;
  for (ullong i = 0; i <= s_nDequeStealerThreadCount; i++) {
    received_sum += sums[i];
    overall_count += succ_counts[i];
  }
  if (overall_count != push_count || received_sum != push_sum) {
    cout << "Incorrect deque\n";
    cout << "Push sum: " << push_sum << "\n";
    cout << "Received sum:" << received_sum << "\n";
    cout << "Push count=" << push_count << "\n";
    cout << "Received count:" << overall_count << "\n";
  }
}

} // namespace
